# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2008-2015
#

import os

import urllib.parse

import logging

from sarracenia import nowflt, timestr2flt, timeflt2str

from sarracenia.flowcb.nodupe import NoDupe

import redis

logger = logging.getLogger(__name__)


class Redis(NoDupe):
    """
    generalised duplicate suppression for sr3 programs. It is used as a 
    time based buffer that prevents, when activated, identical files (of some kinds) 
    from being processed more than once, by rejecting files identified as duplicates.

    options:

    nodupe_ttl - duration in seconds (floating point.)
                The time horizon of the receiption cache.
                how long to remember files, so they are rejected as duplicates.

    The expiry based on nodupe_ttl is applied every housekeeping interval.

    nodupe_fileAgeMax - the oldest file that will be considered for processing.
                        files older than this threshold will be rejected.

    nodupe_fileAgeMin - the newest file that can be considered for processing.
                        files newer than this threshold will be rejected.
                        if not specified, the value of option *inflight*
                        may be referenced if it is an integer value.

    """

    # ----------- magic Methods ------------
    def __init__(self, options):

        super().__init__(options,logger)
        logger.debug("NoDupe_Redis init")
        logging.basicConfig(format=self.o.logFormat, level=getattr(logging, self.o.logLevel.upper()))

        self.o.add_option( 'nodupe_ttl', 'duration', 0 ) 
        self.o.add_option( 'nodupe_fileAgeMax', 'duration', 0 ) 
        self.o.add_option( 'nodupe_fileAgeMin', 'duration', 0 ) 

        logger.info('time_to_live=%d, ' % (self.o.nodupe_ttl))

        self.o.add_option( 'nodupe_redis_serverurl', 'str')
        self.o.add_option( 'nodupe_redis_keybase', 'str', 'sr3.nodupe.' + self.o.component + '.' + self.o.config.replace(".","_")) 

        self._rkey_base = self.o.nodupe_redis_keybase
        self._rkey_count = self._rkey_base + ".count"
        self._cache_hit = None

        self._redis = redis.from_url(self.o.nodupe_redis_serverurl)

        self._last_expire = nowflt()

        self._last_time = nowflt()

        self._redis.set(self._rkey_count, 0, nx=True)
        self._last_count = self._count()


    # ----------- Private Methods -----------
    def _hash(self, text) -> str:
        from hashlib import blake2b
        h = blake2b(key=bytes(self._rkey_base, 'utf-8'), digest_size=16)
        h.update(bytes(text, 'utf-8'))
        return h.hexdigest()
    
    def _is_new(self, message) -> bool :
        """
        Derive keys to be looked up in cache of messages already seen, then look them up in the cache, 

        return False if message is a dupe.
                True if it is new.
        """

        key = self.deriveKey(message)

        if ('nodupe_override' in message) and ('path' in message['nodupe_override']):
            path = message['nodupe_override']['path']
        else:
            path = message['relPath'].lstrip('/')

        message['noDupe'] = { 'key': key, 'path': path }
        message['_deleteOnPost'] |= set(['noDupe'])

        logger.debug("checking (%s, %s)" % (key, path))

        self.cache_hit = None
        key_hashed = self._hash(key)
        path_quoted = urllib.parse.quote(path)
        path_hashed = self._hash(path)

        redis_key = self._rkey_base + ":" + key_hashed + "." + path_hashed

        got = self._redis.get(redis_key)

        #logger.debug("ttl type =%s" % (type(self.o.nodupe_ttl)) )
        self._redis.set(redis_key, str(self.now) + "|" + path_quoted, ex=int(self.o.nodupe_ttl))
        
        if got != None:
            logger.debug("entry already in cache: key=%s" % (redis_key) )
            logger.debug("updated time entry: time=%s" % (str(self.now)) )
            self.cache_hit = path_quoted
            return False
        else:
            logger.debug("adding entry to cache; key=%s" % (redis_key) )
            #self._redis.incr(self._rkey_count)
            return True

    def _count(self):
        count = self._redis.get(self._rkey_count)
        if count == None:
            return 0
        else:
            return int(count)
        
    # ----------- Public Methods -----------
    def on_housekeeping(self):

        logger.info("start")

        new_count = len(self._redis.keys(self._rkey_base + ":*"))
        self.now = nowflt()
        
        logger.info("cache size was %d items %5.2f sec ago, now saved %d entries" % (self._last_count, self.now - self._last_time, new_count))

        self._last_time = self.now

        self._last_count = new_count

    def after_accept(self, worklist):
        new_incoming = []
        self.now = nowflt()
        if self.o.nodupe_fileAgeMax > 0:
            min_mtime = self.now - self.o.nodupe_fileAgeMax
        else:
            min_mtime = 0

        if self.o.nodupe_fileAgeMin > 0:
            max_mtime = self.now - self.o.nodupe_fileAgeMin
        elif type(self.o.inflight) in [ int, float ] and self.o.inflight > 0:
            max_mtime = self.now - self.o.inflight
        else:
            # FIXME: should we add some time here to allow for different clocks?
            #        100 seconds in the future? hmm...
            max_mtime = self.now + 100

        for m in worklist.incoming:
            if ('mtime' in m) :
                mtime=timestr2flt(m['mtime'])
                if mtime < min_mtime:
                    m['_deleteOnPost'] |= set(['reject'])
                    m['reject'] = f"{m['mtime']} too old (nodupe check), oldest allowed {timeflt2str(min_mtime)}"
                    m.setReport(304,  f"{m['mtime']} too old (nodupe check), oldest allowed {timeflt2str(min_mtime)}" )
                    worklist.rejected.append(m)
                    continue
                elif mtime > max_mtime:
                    m['_deleteOnPost'] |= set(['reject'])
                    m['reject'] = f"{m['mtime']} too new (nodupe check), newest allowed {timeflt2str(max_mtime)}"
                    m.setReport(304,  f"{m['mtime']} too new (nodupe check), newest allowed {timeflt2str(max_mtime)}" )
                    worklist.rejected.append(m)
                    continue

            if self._is_new(m):
                new_incoming.append(m)
            else:
                m['_deleteOnPost'] |= set(['reject'])
                m['reject'] = "not modifified 1 (nodupe check)"
                m.setReport(304, 'Not modified 1 (cache check)')
                worklist.rejected.append(m)

        logger.debug("items registered in duplicate suppression cache: %d" % (len(self._redis.keys(self._rkey_base + ":*"))) )
        worklist.incoming = new_incoming

    def on_start(self):
        self._last_count = len(self._redis.keys(self._rkey_base + ":*"))
        

    def on_stop(self):
        self._last_count = None




"""
   Task: look at the fields in the message, and perhaps settings and
         return a new file name for the target of the send or download.

kind of a last resort function, exists mostly for sundew compatibility.
can be used for selective renaming using accept clauses.

The flowcb modules here are for compatiblity with sundew "filename" options.
a module that is focused on changing the names of files being sent or downloaded
is called a Destination File Name script ... or DESTFNSCRIPT 

These are floscb modules that are focused on the destfn entry point.

def destfn( self, msg ):

It's only argument is msg, the message containing the fields to modify.


destfn plugin script is used by senders or subscribers to do complex file naming.
one can invoke a destfn using the *filename* option in a configuration::

     filename DESTFNSCRIPT=sarracenia.flowcb.destfn.sample.Sample

An alternative method of invocation is to apply it selectively in an *accept* line::

     accept k* DESTFNSCRIPT=destfn.sample

As with other flowcb plugins, the import will be done using normal
python import mechanism equivalent to:

     import sarracenia.flowcb.destfn.sample

The destfn routine consults the fields in the given message, and based on them,
return a new file name for the file to have after transfer (download or send.)

the routines have access to the settings via options provided to init,
accessed, by convention, as self.o.

The routine can also modify fields create new ones in the message.

the destfn routine returns the new name of the file.

"""

pass



"""
Posts any new emails from an email server, connected to using 
the specified protocol, either pop3 or imap. The imaplib/poplib 
implementations in Python use the most secure SSL settings by 
default: PROTOCOL_TLS, OP_NO_SSLv2, and OP_NO_SSLv3.
Compatible with Python 2.7+.

A sample do_poll option for sr_poll.
connects to an email server with the provided
credentials and posts all new messages by their msg ID.

usage:
        in an sr_poll configuration file:

        pollUrl [imap|imaps|pop|pops]://[user[:password]@]host[:port]/

        IMAP over SSL uses 993, POP3 over SSL uses 995
        IMAP unsecured uses 143, POP3 unsecured uses 110

        Full credentials must be in credentials.conf.
        If port is not specified it'll default to the ones above based on protocol/ssl setting.

This posts what messages are available. A separate component is needed to 
download the message, which would need:

     callback download.mail_ingest
 
to process these posts.


"""

import datetime
import email
import imaplib
import logging
import poplib
import sarracenia
from sarracenia.flowcb.poll import Poll

logger = logging.getLogger(__name__)


class Mail(Poll):
    def __init__(self, options):

        self.o = options
        logger.info("poll_email_ingest init")

    def poll(self):

        logger.debug("start")

        ok, details = self.o.credentials.get(self.o.pollUrl)
        if ok:
            setting = details.url
            user = setting.username
            password = setting.password
            server = setting.hostname
            protocol = setting.scheme.lower()
            port = setting.port
            logger.debug("pollUrl valid")
        else:
            logger.error("pollUrl: invalid credentials")
            return

        if not port:
            if protocol == "imaps":
                port = 993
            elif protocol == "pops":
                port = 995
            elif protocol == "imap":
                port = 143
            else:
                port = 110

        gathered_messages = []
        if "imap" in protocol:
            if protocol == "imaps":
                try:
                    mailman = imaplib.IMAP4_SSL(server, port=port)
                    mailman.login(user, password)
                except imaplib.IMAP4.error as e:
                    logger.error(
                        "poll_email_ingest imaplib connection error: {}".
                        format(e))
                    return

            elif protocol == "imap":
                try:
                    mailman = imaplib.IMAP4(server, port=port)
                    mailman.login(user, password)
                except imaplib.IMAP4.error as e:
                    logger.error(
                        "poll_email_ingest imaplib connection error: {}".
                        format(e))
                    return
            else:
                return
            # only retrieves unread mail from inbox, change these values as to your preference
            mailman.select(mailbox='INBOX')
            resp, data = mailman.search(None, '(UNSEEN)')
            self.metrics['transferRxBytes'] += len(data)
            for index in data[0].split():
                r, d = mailman.fetch(index, '(RFC822)')
                msg = d[0][1].decode("utf-8", "ignore") + "\n"
                msg_subject = email.message_from_string(msg).get('Subject')
                msg_filename = msg_subject + datetime.datetime.now().strftime(
                    '%Y%m%d_%H%M%s_%f')
                m = sarracenia.Message.fromFileInfo(msg_filename, self.o)
                gathered_messages.append(m)

            mailman.close()
            mailman.logout()

        elif "pop" in protocol:
            if protocol == "pops":
                try:
                    mailman = poplib.POP3_SSL(server, port=port)
                    mailman.user(user)
                    mailman.pass_(password)
                    logger.debug("poll_email_ingest connection started")
                except poplib.error_proto as e:
                    logger.error(
                        "poll_email_ingest pop3 connection error: {}".format(
                            e))
                    return

            elif protocol == "pop":
                try:
                    mailman = poplib.POP3(server, port=port)
                    mailman.user(user)
                    mailman.pass_(password)
                except poplib.error_proto as e:
                    logger.error(
                        "poll_email_ingest pop3 connection error: {}".format(
                            e))
                    return
            else:
                return
            # only retrieves msgs that haven't triggered internal pop3 'read' flag
            numMsgs = len(mailman.list()[1])
            for index in range(numMsgs):
                msg = ""
                for line in mailman.retr(index + 1)[1]:
                    self.metrics['transferRxBytes'] += len(line)
                    msg += line.decode("utf-8", "ignore") + "\n"
                msg_subject = email.message_from_string(msg).get('Subject')
                msg_filename = msg_subject + datetime.datetime.now().strftime(
                    '%Y%m%d_%H%M%s_%f')
                m = sarracenia.Message.fromFileInfo(msg_filename, self.o)
                gathered_messages.append(m)

            mailman.quit()

        else:
            logger.error(
                "poll_email_ingest pollUrl protocol must be one of 'imap/imaps' or 'pop/pops'."
            )
        return gathered_messages



""" msg_pclean_f90 module: second file propagation test for Sarracenia components (in flow test)
"""

from difflib import Differ
import filecmp
import logging
import os
import random

from sarracenia.flowcb.pclean import PClean
from sarracenia import timestr2flt, nowflt

logger = logging.getLogger(__name__)


class PClean_F92(PClean):
    """ This plugin that manage the removal of every file

     - it fails if one removal failed
    """
    def after_accept(self, worklist):
        import os

        logger.info("start len(worklist.incoming) = %d" %
                    len(worklist.incoming))

        outgoing = []

        for msg in worklist.incoming:
            result = True
            ext = self.get_extension('/' + msg['relPath'])
            logger.info("relPath=%s ext: %s in %s ?" %
                        (msg['relPath'], ext, self.test_extension_list))

            if ext in self.test_extension_list:
                f20_path = '/' + msg['relPath'].replace(
                    "{}/".format(self.all_fxx_dirs[1]), self.all_fxx_dirs[0])
                f20_path = f20_path.replace(ext, '')
                try:
                    os.unlink(f20_path)
                    logger.info("unlinked 1: %s" % f20_path)
                except FileNotFoundError as err:
                    logger.error("could not unlink 1 in {}: {}".format(
                        f20_path, err))
                    logger.debug("Exception details:", exc_info=True)
                    result = False
                fxx_dirs = self.all_fxx_dirs[1:2] + self.all_fxx_dirs[6:]
                path_dict = self.build_path_dict(fxx_dirs,
                                                 '/' + msg['relPath'])
                for fxx_dir, path in path_dict.items():
                    try:
                        os.unlink(path)
                        logger.info("unlinked 2: %s" % path)
                        if ext != '.moved':
                            os.unlink(path.replace(ext, ''))
                            logger.info("unlinked 3: %s" %
                                        path.replace(ext, ''))
                    except OSError as err:
                        logger.error("could not unlink 2or3 in {}: {}".format(
                            fxx_dir, err))
                        logger.debug("Exception details:", exc_info=True)
                        result = False
            if result:
                logger.debug('passing to pclean_f92')
                outgoing.append(msg)
            else:
                worklist.rejected.append(msg)

        worklist.incoming = outgoing
        logger.info("len(worklist.incoming) = %d" % len(worklist.incoming))
        logger.info("len(worklist.rejected) = %d" % len(worklist.rejected))



import codecs
import itertools
import os
import re
import sys

from setuptools import find_packages
from distutils.core import setup

#import sarracenia

here = os.path.abspath(os.path.dirname(__file__))


def latest_changelog() -> str:
    
    lines=[]
    started=False
    with open('debian/changelog','r') as clf:
       for l in clf.readlines():
          if len(l) < 2: continue
          if l[2] == '*' :
             lines.append(l)  
          elif l[2] == '-': break
    return '\nChanges:\n' + '\n'.join(lines)

def read(*parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    return codecs.open(os.path.join(here, *parts), 'r').read()

metadata = {}
with open(os.path.join(here, "sarracenia", "_version.py"), "r") as f:
    exec(f.read(), metadata)

packages = find_packages()
print("packages = %s" % packages)

features = {
       'amqp' : [ "amqp" ],
       'filetypes': [ "python-magic" ], 
       'ftppoll' : ['dateparser' ],
       'mqtt': [ 'paho.mqtt>=1.5.1' ],
       'vip': [ 'netifaces' ],
       'redis': [ 'redis' ]
    } 



features['all'] = list(itertools.chain.from_iterable(features.values()))

setup(
    name='metpx-sr3',
    python_requires='>=3.6',
    version=metadata["__version__"],
    description='Subscribe, Acquire, and Re-Advertise products.',
    long_description_content_type='text/x-rst',
    long_description=(read('README.rst')+latest_changelog()), 
    url='https://github.com/MetPX/sarracenia',
    license='GPLv2',
    author='Shared Services Canada, Supercomputing, Data Interchange',
    author_email='Peter.Silva@canada.ca',
    packages=find_packages(),
    package_data={'sarracenia': ['examples/*/*']},
    entry_points={
        "console_scripts": [
            "sr3=sarracenia.sr:main",
            "sr3_post=sarracenia.sr_post:main",
            "sr3_rotateLogsManually=sarracenia.sr_rotateLogsManually:main", 
            #"sr3_poll=sarracenia.sr_flow:main",
            #"sr3_report=sarracenia.sr_flow:main",
            #"sr3_watch=sarracenia.sr_flow:main",
            #"sr3_winnow=sarracenia.sr_flow:main",
            #"sr3_sarra=sarracenia.sr_flow:main",
            #"sr3_shovel=sarracenia.sr_flow:main",
            #"sr3_sender=sarracenia.sr_flow:main",
            #"sr3_subscribe=sarracenia.sr_flow:main",
            #"sr3_log2save=sarracenia.sr_log2save:main",
            "sr3_tailf=sarracenia.sr_tailf:main"
        ]
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Communications :: File Sharing',
        'Topic :: System :: Logging',
    ],
    install_requires=[
        "appdirs", "humanfriendly", "humanize", "jsonpickle", "psutil>=5.3.0", 
        "paramiko", "watchdog",
        'xattr ; sys_platform!="win32"', 'python-magic; sys_platform!="win32"',
        'python-magic-bin; sys_platform=="win32"'
    ],
    # when building on HPC redhat, python OS packages don't exist. 
    # remove the last three lines of install_requires above, aka: 
    # paramiko, watchdog, xattr, python-magic before you do:
    # 
    #    python3 setup.py bdist_rpm

    extras_require = features
    )








"""
   A destfn script to do string replacement.

Sample usage in a config file:

    destfn_replace 1stStringToReplace,replacementOf1st
    destfn_replace SRCN,sssssRCCCn

    accept .*SRCN.* DESTFNSCRIPT=destfn.replace

    # files that do not match 
    accept .*



"""

import logging

from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)



class Replace(FlowCB):

      def __init__(self,options):
          super().__init__(options,logger)
          self.o.add_option('destfn_replace','list' )

      def destfn(self,msg) -> str:

          old_name=msg["relPath"].split('/')[-1]
          new_name=old_name
          for r in self.o.destfn_replace:
               (before,after) = r.split(',')
               new_name = new_name.replace( before, after, 1)
          logger.info('from: %s,  to: %s' % (old_name, new_name ))
 
          return new_name




"""
Plugin renamewhatfn.py:
    This plugin is no longer needed.  Sundew compoatibility was added to Sarracenia, 
    so now can get the same effect by using the *filename* option which works like it
    does in Sundew:

    filename WHATFN

    what it was used for:
    This renamer strips everything from the first colon in the file name to the end.
    This does the same thing as a 'WHATFN' config on a sundew sender.

Example:
    takes px name     : /apps/dms/dms-metadata-updater/data/international_surface/import/mdicp4d:pull-international-metadata:CMC:DICTIONARY:4:ASCII:20160223124648
    rename for        : /apps/dms/dms-metadata-updater/data/international_surface/import/mdicp4d

Usage:
    flowcb sarracenia.flowcb.accept.renamewhatfn.RenameWhatFn

"""

import logging
import sys, os, os.path, time, stat
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class RenameWhatFn(FlowCB):
    def __init__(self, options):
        super().__init__(options,logger)

    def after_accept(self, worklist):
        for message in worklist.incoming:
            parts = message['new_file'].split(':')

            # join mets les ':' entre les parts... donc ajout de ':' au debut
            extra = ':' + ':'.join(parts[1:])
            message['new_file'] = message['new_file'].replace(extra, '')
            message['rename'] = message['rename'].replace(extra, '')



import logging
import os
from sarracenia.flowcb import FlowCB
import time


logger = logging.getLogger(__name__)


class Delete(FlowCB):
    """
    
    callback flowcp.work.delete

        delete working files after download.
        options:
    
        delete_source -- delete the original input file that was present initially and transformed.
        delete_destination -- delete the downloaded file (presumably after other processing has occurred.)
    
    Usage:
        callback work.delete
    
    """
    def __init__(self, options):
        super().__init__(options,logger)
        logger.debug("initialized")
        self.o.add_option('delete_source', 'flag', True)
        self.o.add_option('delete_destination', 'flag', False)
        self.dirsOfDeletion=set([])

         # dirs not to be deleted.
        self.sacredDirs=set([])
        if hasattr(self.o, 'baseDir'):
            self.sacredDirs.add(self.o.baseDir)
        if hasattr(self.o, 'post_baseDir'):
            self.sacredDirs.add(self.o.post_baseDir)


    def after_accept(self, worklist):
        new_incoming = []
        if self.o.delete_source:
            for message in worklist.incoming:
                message['_deleteOnPost'] |= set(['delete_source'])

                #FIXME: there should be some reference to baseDir here... for url's that aren't file ones.
                #   just getting it to work for particular case for now 2021/12/09 - pas
                message['delete_source'] = message['baseUrl'].lstrip(
                    'file:') + message['relPath']

    def after_work(self, worklist):

        for message in worklist.ok:

            # don't remove files that have already been removed.
            if 'fileOp' in message and 'remove' in message['fileOp']:
                continue

            files_to_delete = []

            if self.o.delete_source:
                files_to_delete.append(message['delete_source'])

            if self.o.delete_destination:
                self.dirsOfDeletion |= set(message['new_dir'])
                files_to_delete.append(
                    "%s/%s" % (message['new_dir'], message['new_file']))

            for f in files_to_delete:
                logger.info("deleting %s" % f)
                try:
                    os.unlink(f)
                except OSError as err:
                    logger.error("could not unlink {}: {}".format(f, err))
                    logger.debug("Exception details:", exc_info=True)

    def on_housekeeping(self):

        dirlist=self.dirsOfDeletion

        logger.info('scan for directories to cleanup')
        for d in dirlist:
            if d in self.sacredDirs:
                self.dirsOfDeletion.remove(d)
                continue
            if os.path.isdir(d):
                l = os.listdir(d)
                if len(l) == 0: # empty directory
                    logger.info( f"found {d} is empty.")
                    s = os.stat(d)
                    age = time.time() - s.st_mtime
                    if age > self.o.housekeeping:
                        try:
                            os.rmdir(d)
                            self.dirsOfDeletion.remove(d)
                            self.dirsofDeltion.add(dirname(d))
                            logger.info( f"deleted {d}")
                        except Exception as err:
                            logger.error("could not unlink {}: {}".format(f, err))
                            logger.debug("Exception details:", exc_info=True)
                    else:
                        logger.info( f"but not for long enough yet.")




import re
import sarracenia.config
from sarracenia.flow.subscribe import Subscribe
import sarracenia.flowcb
import sarracenia.credentials

cfg = sarracenia.config.no_file_config()

cfg.broker = sarracenia.credentials.Credential(
    'amqps://anonymous:anonymous@hpfx.collab.science.gc.ca')
cfg.topicPrefix = ['v02', 'post']
cfg.component = 'subscribe'
cfg.config = 'flow_demo'
cfg.bindings = [('xpublic', ['v02', 'post'],
                 ['*', 'WXO-DD', 'observations', 'swob-ml', '#'])]
cfg.queueName = 'q_anonymous.subscriber_test2'
cfg.download = True
cfg.batch = 1
cfg.messageCountMax = 5

# set the instance number for the flow class.
cfg.no = 0

# set other settings based on what is provided, for things like state files.
cfg.finalize()

# accept/reject patterns:
pattern = ".*"
#              to_match, write_to_dir, DESTFN, regex_to_match, accept=True,mirror,strip, pstrip,flatten
cfg.masks = [(pattern, "/tmp/flow_demo", None, re.compile(pattern), True,
              False, False, False, '/')]

subscriber = sarracenia.flow.subscribe.Subscribe(cfg)

subscriber.run()



import logging
import sarracenia.flowcb

logger = logging.getLogger(__name__)


class Sample(sarracenia.flowcb.FlowCB):
    def __init__(self, options):

        super().__init__(options,logger)

        # declare a module specific setting.
        options.add_option('announce_list', 'list')

    def on_start(self):

        logger.info('announce_list: %s' % self.o.announce_list)



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2008-2015
#
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://github.com/MetPX/sarracenia
#
########################################################################
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#
#

import calendar, datetime
from hashlib import md5
from hashlib import sha512
import logging
import os
import random
import signal
import stat
import sys
import time
import urllib
import urllib.parse

#from sarracenia.sr_xattr import *
from sarracenia import nowflt, timestr2flt
from sarracenia.featuredetection import features

logger = logging.getLogger(__name__)

#============================================================
# sigalarm
#============================================================


class TimeoutException(Exception):
    """timeout exception"""
    pass


# alarm_cancel
def alarm_cancel():
    if sys.platform != 'win32':
        signal.alarm(0)


# alarm_raise
def alarm_raise(n, f):
    raise TimeoutException("signal alarm timed out")


# alarm_set
def alarm_set(time):
    """
       FIXME: replace with set itimer for > 1 second resolution... currently rouding to nearest second. 
    """

    if sys.platform != 'win32':
        signal.signal(signal.SIGALRM, alarm_raise)
        signal.alarm(int(time + 0.5))


# =========================================
# sr_proto : one place for throttle, onfly checksum, buffer io timeout
# =========================================


class Transfer():
    """
     This is a sort of abstract base classe for implementing transfer protocols.
     Implemented subclasses include support for: local files, https, sftp, and ftp. 

     This class has routines that do i/o given descriptors opened by the sub-classes,
     so that each one does not need to re-implement copying, for example.

     Each subclass needs to implement the following routines:

     if downloading:: 

         get    ( msg, remote_file, local_file, remote_offset=0, local_offset=0, length=0 )
         getAccellerated( msg, remote_file, local_file, length ) 
         ls     ()
         cd     (dir)
         delete (path)
    
     if sending:: 

         put    ( msg, remote_file, local_file, remote_offset=0, local_offset=0, length=0 )
         putAccelerated ( msg, remote_file, local_file, length=0 )
         cd     (dir)
         mkdir  (dir)
         umask  ()
         chmod  (perm)
         rename (old,new)
    
     uses options (on Sarracenia.config data structure passed to constructor/factory.)
     * credentials - used to authentication information.
     * sendTo  - server to connect to.
     * batch   - how many files to transfer before a connection is torn down and re-established.
     * permDefault - what permissions to set on files transferred.
     * permDirDefault - what permission to set on directories created.
     * timeout  - how long to wait for operations to complete.
     * byteRateMax - maximum transfer rate (throttle to avoid exceeding)
     * bufsize - size of buffers for file transfers.

    """
    @staticmethod
    def factory(proto, options) -> 'Transfer':

        for sc in Transfer.__subclasses__():
            if (hasattr(sc, 'registered_as')
                    and (proto in sc.registered_as())):
                return sc(proto, options)
        return None

    def __init__(self, proto, options):

        self.o = options
        if 'sarracenia.transfer.Transfer' in self.o.settings and 'logLevel' in self.o.settings:
            ll = self.o.settings['sarracenia.transfer.Transfer']['logLevel']
        else:
            ll = self.o.logLevel

        logger.setLevel(getattr(logging, ll.upper()))

        logger.debug("class=%s , subclasses=%s" %
                     (type(self).__name__, Transfer.__subclasses__()))
        self.init()

    def init(self):
        self.sumalgo = None
        self.checksum = None
        self.data_sumalgo = None
        self.data_checksum = None
        self.fpos = 0
        self.tbytes = 0
        self.tbegin = nowflt()
        self.byteRate = 0

    def local_read_close(self, src):
        #logger.debug("sr_proto local_read_close")

        src.close()

        # finalize checksum

        if self.sumalgo: self.checksum = self.sumalgo.value
        if self.data_sumalgo:
            self.data_checksum = self.data_sumalgo.value

    def local_read_open(self, local_file, local_offset=0):
        logger.debug("sr_proto local_read_open getcwd=%s self.cwd=%s" %
                     (os.getcwd(), self.getcwd()))

        self.checksum = None

        # local_file opening and seeking if needed

        src = open(local_file, 'rb')
        if local_offset != 0: src.seek(local_offset, 0)

        # initialize sumalgo

        if self.sumalgo: self.sumalgo.set_path(local_file)
        if self.data_sumalgo: self.data_sumalgo.set_path(local_file)

        return src

    def local_write_close(self, dst):

        # flush sync (make sure all io done)

        dst.flush()
        os.fsync(dst)

        # flush,sync, remember current position, truncate = no sparce, close

        self.fpos = dst.tell()
        dst.truncate()
        dst.close()

        # finalize checksum

        if self.sumalgo: self.checksum = self.sumalgo.value
        if self.data_sumalgo:
            self.data_checksum = self.data_sumalgo.value

    def local_write_open(self, local_file, local_offset=0):
        #logger.debug("sr_proto local_write_open")

        # reset ckecksum, fpos

        self.checksum = None
        self.fpos = 0

        # local_file has to exists

        if not os.path.isfile(local_file):
            dst = open(local_file, 'w')
            dst.close()

        # local_file opening and seeking if needed

        dst = open(local_file, 'r+b')
        if local_offset != 0: dst.seek(local_offset, 0)

        return dst

    def on_data(self, chunk) -> bytes:
        """
            transform data as it is being read. 
            Given a buffer, return the transformed buffer. 
            Checksum calculation is based on pre transformation... likely need
            a post transformation value as well.
        """

        return chunk

        #FIXME ... need to re-enable on_data plugins... not sure how they should work.
        # sub-classing of transfer class?

    def read_write(self, src, dst, length=0):
        logger.debug("sr_proto read_write")

        # reset speed

        rw_length = 0
        self.tbytes = 0.0
        self.tbegin = nowflt()

        # length = 0, transfer entire remote file to local file

        if length == 0:
            while True:
                if self.o.timeout: alarm_set(self.o.timeout)
                chunk = src.read(self.o.bufsize)
                if chunk:
                    new_chunk = self.on_data(chunk)
                    dst.write(new_chunk)
                    rw_length += len(chunk)
                alarm_cancel()
                if not chunk: break
                if self.sumalgo: self.sumalgo.update(chunk)
                self.throttle(chunk)
            return rw_length

        # exact length to be transfered

        nc = int(length / self.o.bufsize)
        r = length % self.o.bufsize

        # read/write bufsize "nc" times

        i = 0
        while i < nc:
            if self.o.timeout: alarm_set(self.o.timeout)
            chunk = src.read(self.o.bufsize)
            if chunk:
                new_chunk = self.on_data(chunk)
                rw_length += len(new_chunk)
                dst.write(new_chunk)
            alarm_cancel()
            if not chunk: break
            if self.sumalgo: self.sumalgo.update(chunk)
            self.throttle(chunk)
            i = i + 1

        # remaining

        if r > 0:
            if self.o.timeout: alarm_set(self.o.timeout)
            chunk = src.read(r)
            if chunk:
                new_chunk = self.on_data(chunk)
                rw_length += len(new_chunk)
                dst.write(new_chunk)
            alarm_cancel()
            if self.sumalgo: self.sumalgo.update(chunk)
            self.throttle(chunk)

        return rw_length

    def read_writelocal(self,
                        src_path,
                        src,
                        local_file,
                        local_offset=0,
                        length=0):
        #logger.debug("sr_proto read_writelocal")

        # open
        dst = self.local_write_open(local_file, local_offset)

        # initialize sumalgo

        if self.sumalgo: self.sumalgo.set_path(src_path)
        if self.data_sumalgo: self.data_sumalgo.set_path(src_path)

        # copy source to sendTo

        # 2022/12/02 - pas - need copies to always work...
        # in HPC mirroring case, a lot of short files, likely length is wrong in announcements.
        # grab the whole file unconditionally for now, detect error using mismatch.
        rw_length = self.read_write(src, dst, 0)

        # close
        self.local_write_close(dst)

        # warn if length mismatch without transformation.
        # 2022/12/02 - pas should see a lot of these messages in HPC case from now on...
        if length != 0 and rw_length != length:
            logger.warning(
                "util/writelocal mismatched file length writing %s. Message said to expect %d bytes.  Got %d bytes."
                % (local_file, length, rw_length))

        return rw_length

    def readlocal_write(self, local_file, local_offset=0, length=0, dst=None):
        logger.debug("sr_proto readlocal_write")

        # open
        src = self.local_read_open(local_file, local_offset)

        # copy source to sendTo

        rw_length = self.read_write(src, dst, length)

        # close

        self.local_read_close(src)

        # warn if length mismatch without transformation.

        # FIXME: 2020/09 - commented out for now... unsure about this.
        #if (not self.o.on_data_list) and length != 0 and rw_length != length :
        #   logger.error("util/readlocal mismatched file length reading %s. Message announced it as %d bytes, but read %d bytes " % (local_file,length,rw_length))

        # 2022/12/02 - pas attempting to get files that get shorter addressed.
        if ((length==0) or (rw_length < length)) and hasattr(dst,'truncate'):
             dst.truncate(rw_length)

        return rw_length

    def set_sumalgo(self, sumalgo):
        logger.debug("sr_proto set_sumalgo %s" % sumalgo)
        self.sumalgo = sarracenia.identity.Identity.factory(sumalgo)
        self.data_sumalgo = sarracenia.identity.Identity.factory(sumalgo)

    def set_sumArbitrary(self, value):
        self.sumalgo.value = value
        self.data_sumalgo.value = value

    def update_file(self, path):
        if self.sumalgo:
            self.sumalgo.update_file(path)
        if self.data_sumalgo:
            self.data_sumalgo.update_file(path)

    def set_path(self, path):
        if self.sumalgo:
            self.sumalgo.set_path(path)
        if self.data_sumalgo:
            self.data_sumalgo.set_path(path)

    def get_sumstr(self) -> dict:
        if self.sumalgo:
            #return { 'method':type(self.sumalgo).__name__, 'value':self.sumalgo.value }
            return {
                'method': self.sumalgo.get_method(),
                'value': self.sumalgo.value
            }
        else:
            return None

    def metricsReport(self):
        return { 'byteRateInstant': self.byteRate }

    # throttle
    def throttle(self, buf):
        self.tbytes = self.tbytes + len(buf)
        rspan = nowflt() - self.tbegin
        if rspan > 0:
            self.byteRate = self.tbytes/rspan

        if hasattr(self.o,'byteRateMax') and self.o.byteRateMax and self.o.byteRateMax > 0:
            span = self.tbytes / self.o.byteRateMax
            if span > rspan:
                stime = span - rspan
                if stime > 10:
                    logger.info( f"exceeded byteRateMax: {self.o.byteRateMax} sleeping for {stime:.2f}")
                time.sleep(stime)

    # write_chunk
    def write_chunk(self, chunk):
        if self.chunk_iow: self.chunk_iow.write(chunk)
        self.rw_length += len(chunk)
        alarm_cancel()
        if self.sumalgo: self.sumalgo.update(chunk)
        self.throttle(chunk)
        if self.o.timeout: alarm_set(self.o.timeout)

    # write_chunk_end
    def write_chunk_end(self):
        alarm_cancel()
        self.chunk_iow = None
        return self.rw_length

    # write_chunk_init
    def write_chunk_init(self, proto):
        self.chunk_iow = proto
        self.tbytes = 0.0
        self.tbegin = nowflt()
        self.rw_length = 0
        if self.o.timeout: alarm_set(self.o.timeout)

# batteries included.
import sarracenia.transfer.file
import sarracenia.transfer.ftp
import sarracenia.transfer.https

if features['sftp']['present']:
    import sarracenia.transfer.sftp




from sarracenia.flow import Flow
import logging

logger = logging.getLogger(__name__)

default_options = {'acceptUnmatched': True, 'download': True}


class Sarra(Flow):
    """
       * download files from a remote server to the local one
       * modify the messages so they refer to the downloaded files.
       * re-post them to another exchange for the next other subscribers.
    """
    def __init__(self, options):

        super().__init__(options)
        self.plugins['load'].insert(
            0, 'sarracenia.flowcb.gather.message.Message')

        if hasattr(self.o, 'post_exchange'):
            self.plugins['load'].insert(
                0, 'sarracenia.flowcb.post.message.Message')

    def do(self):

        if self.o.download:
            self.do_download()
        else:
            # mark all remaining messages as done.
            self.worklist.ok = self.worklist.incoming
            self.worklist.incoming = []

        logger.debug('processing %d messages worked!' % len(self.worklist.ok))



import json
import logging
import sarracenia
from sarracenia.postformat import PostFormat
import urllib
import uuid

#from pywis_pubsub.publish import create_message


logger = logging.getLogger(__name__)


class Wis(PostFormat):
    """

       WMO Information Service

       WMO - World Meteorological Organization

       a standard in development for exchange of weather data.

       A class for controlling the format of messages are sent.
       internally All messages are represented as v03. 
       post format implement translations to other protocols for interop.

       STATUS: creates bogus, but "correctly" formatted messages
           not working... can't even figure out what the WMO wants... 

       * there is no algorithm for the topic yet. just post a bogus one.
       * data-id field... no algorithm to derive it yet... uuid was once suggested.
       * geometry field should be set before calling.
       
       from internal format:
       * sarracenia does not include the content-type of the data in the message, so that is omitted for now.
         (reading up on relevant mime-type standards, when you do not know, you are supposed to omit.)


   """

    @staticmethod
    def content_type():
        return 'application/geo+json'

    @staticmethod
    def mine(payload, headers, content_type, options) -> bool:
        """
          return true if the message is in this encoding.
       """
        if content_type == Wis.content_type():
            return True
        return False

    @staticmethod
    def importMine(body, headers, options) -> sarracenia.Message:
            """
          given a message in a wire format, with the given properties (or headers) in a dictionary,
          return the message as a normalized v03 message.

          Since I can't figure out a baseUrl, I can't import any messages. so import is broken for now.
          for now:  just parse the first url in 'links', assign the host part to baseUrl, and make
          the path a retrieval one.

       """
            msg = sarracenia.Message()
            msg["_format"] = __name__.split('.')[-1].lower()
            try:
                GeoJSONBody=json.loads(body)
            except Exception as ex:
                logger.warning('expected geojson, decode error: %s' % ex)
                logger.debug('Exception details: ', exc_info=True)
                return None

            if 'properties' in GeoJSONBody:
                if 'pubtime' in GeoJSONBody['properties']: 
                    t=GeoJSONBody['properties']['pubtime']
                    msg['pubTime'] = t[0:4]+t[5:7]+t[8:13]+t[14:16]+t[17:-1]
                else:
                    logger.error( 'invalid message missing pubtime (WMO mandatory field)' )
  
                for h in GeoJSONBody['properties']:
                    if h not in [ 'pubtime' ]:
                        msg[h] = GeoJSONBody['properties'][h]

            #logger.warning( f" headers: {headers}, msg: {msg}  ... GeoJSONBody: {GeoJSONBody}  ")
            if not 'type' in GeoJSONBody:
                logger.warning( 'invalid message missing type (WMO mandatory field)' )

            if 'geometry' in GeoJSONBody :
                if GeoJSONBody['geometry'] is not None:
                    msg['geometry'] = GeoJSONBody['geometry']
            else:
                logger.warning( 'invalid message missing geometry (WMO mandatory field)' )

            if not ( 'version' in GeoJSONBody and GeoJSONBody['version'] == 'v04' ):
                logger.warning( 'invalide message missing version (WMO Mandatory field)' )

            if ('data_id' in msg) and ('topic' in headers):
                msg['relPath'] = headers['topic'] + '/' + msg['data_id']
            else:
                logger.warning( 'invalid message missing data_id (WMO mandatory field)' )
           
            if 'links' in GeoJSONBody:
                urlstr = GeoJSONBody['links'][0]['href']
                url = urllib.parse.urlparse( urlstr )
                msg['size']  = GeoJSONBody['links'][0]['length']
                if 'type' in GeoJSONBody['links'][0]:
                    msg['contentType']  = GeoJSONBody['links'][0]['type']
                msg['links'] = GeoJSONBody['links']
                msg['baseUrl'] = url.scheme + '://' + url.netloc
                msg['retrievePath' ] = urlstr[len(msg['baseUrl']):] 
            else:
                logger.warning( 'message missing links (WMO mandatory field)' )

            return msg

    @staticmethod
    def exportMine(body, options) -> (str, dict, str):
            """
           given a v03 (internal) message, produce an encoded version.
       """
            GeoJSONBody={ 'type': 'Feature', 'geometry': None, 'properties':{}, 'version':'v04' }

            for literal in [ 'geometry', 'properties' ]:
                if literal in body:
                    GeoJSONBody[literal] = body[literal]

            if 'topic' in options:
                topic=options['topic'].split('/')
            else:
                topic= []

            headers = { 'topic' : topic }
            """
                  topicPrefix and body['subtopic'] could be used to build a topic...

            """
            for h in body:
                if h not in [ 'geometry', 'properties', 'size', 'baseUrl', 'relPath', 'retrievePath', 'subtopic', 'pubTime' ]:
                    GeoJSONBody['properties'][h] = body[h]

            t=body['pubTime']
            GeoJSONBody['properties']['pubtime'] = t[0:4]+'-'+t[4:6]+'-'+t[6:11]+':'+t[11:13]+':'+t[13:] + 'Z'

            if 'geometry' in body:
                GeoJSONBody['geometry'] = body['geometry']

            if 'data_id' not in body:
                GeoJSONBody['properties']['data_id'] =  str(uuid.uuid4())

            if 'retrievePath' in body :
                url = body['baseUrl'] + body['retrievePath']
            else:
                url = body['baseUrl'] + body['relPath']

            if not 'links' in GeoJSONBody:

                GeoJSONBody['links'] = [{
                    'rel': 'canonical',
                    #'type': 'unknown',, mime content-type of data, not available in message.
                    'href': url,
                    'length': body['size']
                }]
                if 'contentType' in body:
                    GeoJSONBody['links'][0]['type'] = body['contentType']

            raw_body = json.dumps(GeoJSONBody)
            return raw_body, headers, Wis.content_type()



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, 2008-2020
#
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://github.com/MetPX/sarracenia
#
########################################################################
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#
#

# the real AMQP library... not this one...
import amqp
import copy
import json

import logging

from urllib.parse import unquote
import sarracenia
from sarracenia.postformat import PostFormat
from sarracenia.moth import Moth
import signal

import time

logger = logging.getLogger(__name__)
"""
amqp_ss_maxlen 

the maximum length of a "short string", as per AMQP protocol, in bytes.

"""
amqp_ss_maxlen = 255

default_options = {
    'auto_delete': False,
    'batch': 25,
    'durable': True,
    'exchange': None,
    'exchangeDeclare': True,
    'expire': None,
    'logLevel': 'info',
    'prefetch': 25,
    'queueName': None,
    'queueBind': True,
    'queueDeclare': True,
    'reset': False,
    'subtopic': [],
    'messageDebugDump': False,
    'topicPrefix': ['v03'],
    'vhost': '/',
}


class AMQP(Moth):
    """
       implementation of the Moth API for the amqp library, which is built to talk to rabbitmq brokers in 0.8 and 0.9
       AMQP dialects.

       to allow acknowledgements we map: AMQP' 'delivery_tag' to the 'ack_id'

       additional AMQP specific options:

       exchangeDeclare  - declare exchanges before use.
       queueBind        - bind queue to exchange before use.
       queueDeclare     - declare queue before use.

    """
    def _msgRawToDict(self, raw_msg) -> sarracenia.Message:
        if raw_msg is not None:
            body = raw_msg.body

            if self.o['messageDebugDump']:
                logger.info('raw message start')
                if not ('content_type' in raw_msg.properties):
                    logger.warning('message is missing content-type header')
                if body:
                    logger.info('body: type: %s (%d bytes) %s' %
                             (type(body), len(body), body))
                else:
                    logger.info('had no body')
                if raw_msg.headers:
                    logger.info('headers: type: %s (%d elements) %s' %
                             (type(raw_msg.headers), len(raw_msg.headers), raw_msg.headers))
                else:
                    logger.info('had no headers')
                if raw_msg.properties:
                    logger.info('properties:' % raw_msg.properties)
                else:
                    logger.info('had no properties')
                if raw_msg.delivery_info: 
                    logger.info( f"delivery info: {raw_msg.delivery_info}" )
                else:
                    logger.info('had no delivery info')
                logger.info('raw message end')



            if type(body) is bytes:
                try:
                    body = raw_msg.body.decode("utf8")
                except Exception as ex:
                    logger.error(
                        'ignoring message. UTF8 encoding expected. raw message received: %s' % ex)
                    logger.debug('Exception details: ', exc_info=True)
                    return None

            if 'content_type' in raw_msg.properties:
                content_type = raw_msg.properties['content_type']
            else:
                content_type = None

            msg = PostFormat.importAny( body, raw_msg.headers, content_type, self.o )
            if not msg:
                logger.error('Decode failed, discarding message')
                return None

            topic = raw_msg.delivery_info['routing_key'].replace(
                '%23', '#').replace('%22', '*')
            msg['exchange'] = raw_msg.delivery_info['exchange']
            msg['subtopic'] = topic.split('.')[len(self.o['topicPrefix']):]
            msg['ack_id'] = raw_msg.delivery_info['delivery_tag']
            msg['local_offset'] = 0
            msg['_deleteOnPost'] |= set( ['ack_id', 'exchange', 'local_offset', 'subtopic'])
            if not msg.validate():
                if hasattr(self,channel):
                    self.channel.basic_ack(msg['ack_id'])
                logger.error('message acknowledged and discarded: %s' % msg)
                msg = None
        else:
            msg = None

        return msg

    # length of an AMQP short string (used for headers and many properties)
    amqp_ss_maxlen = 255

    def __init__(self, props, is_subscriber) -> None:
        """
           connect to broker, depending on message_strategy stubborness, remain connected.
           
        """

        super().__init__(props, is_subscriber)

        self.last_qDeclare = time.time()

        logging.basicConfig(
            format=
            '%(asctime)s [%(levelname)s] %(name)s %(funcName)s %(message)s')

        self.o = copy.deepcopy(default_options)
        self.o.update(props)

        self.first_setup = True
        self.please_stop = False

        me = "%s.%s" % (__class__.__module__, __class__.__name__)

        if ('settings' in self.o) and (me in self.o['settings']):
            for s in self.o['settings'][me]:
                self.o[s] = self.o['settings'][me][s]

            if 'logLevel' in self.o['settings'][me]:
                logger.setLevel(self.o['logLevel'].upper())

        self.connection = None

    def __connect(self, broker) -> bool:
        """
          connect to broker. 
          returns True if connected, false otherwise.
            side effect: self.channel set to a new channel.

          Expect caller to handle errors.
        """
        if broker.url.hostname:
            host = broker.url.hostname
            if broker.url.port is None:
                if (broker.url.scheme[-1] == 's'):
                    host += ':5671'
                else:
                    host += ':5672'
            else:
                host += ':{}'.format(broker.url.port)
        else:
            logger.critical( f"invalid broker specification: {broker} " )
            return False

        # if needed, set the vhost using the broker URL's path
        vhost = self.o['vhost']
        # if the URL path is '/' or '', no vhost is specified and the default vhost from self.o
        # will be used. Otherwise, strip off leading or trailing slashes.
        if broker.url.path != '/' and broker.url.path != '':
            vhost = broker.url.path.strip('/')

        self.connection = amqp.Connection(host=host,
                                          userid=broker.url.username,
                                          password=unquote(
                                              broker.url.password),
                                          login_method=broker.login_method,
                                          virtual_host=vhost,
                                          ssl=(broker.url.scheme[-1] == 's'))
        if hasattr(self.connection, 'connect'):
            # check for amqp 1.3.3 and 1.4.9 because connect doesn't exist in those older versions
            self.connection.connect()

        self.channel = self.connection.channel()
        self.management_channel = self.connection.channel()
        return True

    def _amqp_setup_signal_handler(self, signum, stack):
        logger.info("ok, asked to stop")
        self.please_stop=True

    def metricsReport(self):

        if 'no' in self.o and self.o['no'] < 2 and self.is_subscriber and self.connection and self.connection.connected:
            # control frequency of checks for queue size. hardcoded for now.
            next_time = self.last_qDeclare + 30
            now=time.time()
            if next_time <= now:
                self._queueDeclare(passive=True)
                self.last_qDeclare=now

        super().metricsReport()

        return self.metrics

    def _queueDeclare(self,passive=False) ->  int:

        try:
            # from sr_consumer.build_connection...
            if not self.__connect(self.o['broker']):
                logger.critical('could not connect')
                if hasattr(self,'metrics'):
                    self.metrics['brokerQueuedMessageCount'] = -2
                return -2

            #FIXME: test self.first_setup and props['reset']... delete queue...
            broker_str = self.o['broker'].url.geturl().replace(
                ':' + self.o['broker'].url.password + '@', '@')

            if self.o['queueDeclare'] and self.o['queueName']:

                args = {}
                if self.o['expire']:
                    x = int(self.o['expire'] * 1000)
                    if x > 0: args['x-expires'] = x
                if self.o['message_ttl']:
                    x = int(self.o['message_ttl'] * 1000)
                    if x > 0: args['x-message-ttl'] = x

                #FIXME: conver expire, message_ttl to proper units.
                if self.o['dry_run']:
                    logger.info('queue declare (dry run) %s (as: %s) ' %
                            (self.o['queueName'], broker_str))
                    msg_count=0
                else:
                    qname, msg_count, consumer_count = self.management_channel.queue_declare(
                        self.o['queueName'],
                        passive=passive,
                        durable=self.o['durable'],
                        exclusive=False,
                        auto_delete=self.o['auto_delete'],
                        nowait=False,
                        arguments=args)
                    logger.info( f"queue declared {self.o['queueName']} (as: {broker_str}), (messages waiting: {msg_count})" )

                if hasattr(self,'metrics'):
                   self.metrics['brokerQueuedMessageCount'] = msg_count
                return msg_count

        except Exception as err:
            logger.error(
                    f'connecting to: {self.o["queueName"]}, durable: {self.o["durable"]}, expire: {self.o["expire"]}, auto_delete={self.o["auto_delete"]}'
                )
            logger.error("AMQP getSetup failed to {} with {}".format(
                    self.o['broker'].url.hostname, err))
            logger.debug('Exception details: ', exc_info=True)

        if hasattr(self,'metrics'):
            self.metrics['brokerQueuedMessageCount'] = -1
        return -1


    def getSetup(self) -> None:
        """
        Setup so we can get messages.

        if message_strategy is stubborn, will loop here forever.
             connect, declare queue, apply bindings.
        """

        ebo = 1
        original_sigint = signal.getsignal(signal.SIGINT)
        original_sigterm = signal.getsignal(signal.SIGINT)
        signal.signal(signal.SIGINT, self._amqp_setup_signal_handler)
        signal.signal(signal.SIGTERM, self._amqp_setup_signal_handler)

        while True:
            
            if self.please_stop:
                break

            if 'broker' not in self.o or self.o['broker'] is None:
                logger.critical( f"no broker given" )
                break

            # It does not really matter how it fails, the recovery approach is always the same:
            # tear the whole thing down, and start over.
            try:
                # from sr_consumer.build_connection...
                if not self.__connect(self.o['broker']):
                    logger.critical('could not connect')
                    break

                # only first/lead instance needs to declare a queue and bindings.
                if 'no' in self.o and self.o['no'] >= 2:
                    self.metricsConnect()
                    return

                #logger.info('getSetup connected to {}'.format(self.o['broker'].url.hostname) )

                if self.o['prefetch'] != 0:
                    self.channel.basic_qos(0, self.o['prefetch'], True)

                #FIXME: test self.first_setup and props['reset']... delete queue...
                broker_str = self.o['broker'].url.geturl().replace(
                    ':' + self.o['broker'].url.password + '@', '@')

                # from Queue declare
                msg_count = self._queueDeclare()
                
                if msg_count == -2: break

                if self.o['queueBind'] and self.o['queueName']:
                    for tup in self.o['bindings']:
                        exchange, prefix, subtopic = tup
                        topic = '.'.join(prefix + subtopic)
                        if self.o['dry_run']:
                            logger.info('binding (dry run) %s with %s to %s (as: %s)' % \
                                ( self.o['queueName'], topic, exchange, broker_str ) )
                        else:
                            logger.info('binding %s with %s to %s (as: %s)' % \
                                ( self.o['queueName'], topic, exchange, broker_str ) )
                            if exchange:
                                self.management_channel.queue_bind(self.o['queueName'], exchange,
                                                topic)

                # Setup Successfully Complete!
                self.metricsConnect()
                logger.debug('getSetup ... Done!')
                break

            except Exception as err:
                logger.error(
                    f'connecting to: {self.o["queueName"]}, durable: {self.o["durable"]}, expire: {self.o["expire"]}, auto_delete={self.o["auto_delete"]}'
                )
                logger.error("AMQP getSetup failed to {} with {}".format(
                    self.o['broker'].url.hostname, err))
                logger.debug('Exception details: ', exc_info=True)

            if not self.o['message_strategy']['stubborn']: return

            if ebo < 60: ebo *= 2

            logger.info("Sleeping {} seconds ...".format(ebo))
            time.sleep(ebo)

        signal.signal(signal.SIGINT, original_sigint)
        signal.signal(signal.SIGTERM, original_sigterm)
        if self.please_stop:
            signal.raise_signal(signal.SIGINT)

    def putSetup(self) -> None:

        ebo = 1
        original_sigint = signal.getsignal(signal.SIGINT)
        original_sigterm = signal.getsignal(signal.SIGINT)
        signal.signal(signal.SIGINT, self._amqp_setup_signal_handler)
        signal.signal(signal.SIGTERM, self._amqp_setup_signal_handler)

        while True:

            # It does not really matter how it fails, the recovery approach is always the same:
            # tear the whole thing down, and start over.
            try:
                if self.please_stop:
                    break

                if self.o['broker'] is None:
                    logger.critical( f"no broker given" )
                    break

                if not self.__connect(self.o['broker']):
                    logger.critical('could not connect')
                    break

                # transaction mode... confirms would be better...
                self.channel.tx_select()
                broker_str = self.o['broker'].url.geturl().replace(
                    ':' + self.o['broker'].url.password + '@', '@')

                #logger.debug('putSetup ... 1. connected to {}'.format(broker_str ) )

                if self.o['exchangeDeclare']:
                    logger.debug('putSetup ... 1. declaring {}'.format(
                        self.o['exchange']))
                    if type(self.o['exchange']) is not list:
                        self.o['exchange'] = [self.o['exchange']]
                    for x in self.o['exchange']:
                        if self.o['dry_run']:
                            logger.info('exchange declare (dry run): %s (as: %s)' %
                                    (x, broker_str))
                        else:
                            self.channel.exchange_declare(
                                x,
                                'topic',
                                auto_delete=self.o['auto_delete'],
                                durable=self.o['durable'])
                            logger.info('exchange declared: %s (as: %s)' %
                                    (x, broker_str))

                # Setup Successfully Complete!
                self.metricsConnect()
                logger.debug('putSetup ... Done!')
                break

            except Exception as err:
                logger.error(
                    "AMQP putSetup failed to connect or declare exchanges {}@{} on {}: {}"
                    .format(self.o['exchange'], self.o['broker'].url.username,
                            self.o['broker'].url.hostname, err))
                logger.debug('Exception details: ', exc_info=True)

            if not self.o['message_strategy']['stubborn']: return

            if ebo < 60: ebo *= 2

            self.close()
            logger.info("Sleeping {} seconds ...".format(ebo))
            time.sleep(ebo)

        signal.signal(signal.SIGINT, original_sigint)
        signal.signal(signal.SIGTERM, original_sigterm)
        if self.please_stop:
            signal.raise_signal(signal.SIGINT)


    def putCleanUp(self) -> None:

        try:
            for x in self.o['exchange']:
                try:
                    if self.o['dry_run']:
                        logger.info("deleted exchange (dry run): %s (if unused)" % x)
                    else:
                        if hasattr(self,'channel'):
                            self.channel.exchange_delete(x, if_unused=True)
                        logger.info("deleted exchange: %s" % x)
                except amqp.exceptions.PreconditionFailed as err:
                    err_msg = str(err).replace("Exchange.delete: (406) PRECONDITION_FAILED - exchange ", "")
                    logger.warning("failed to delete exchange: %s" % err_msg)
        except Exception as err:
            logger.error("failed on {} with {}".format(
                self.o['broker'].url.hostname, err))
            logger.debug('Exception details: ', exc_info=True)

    def getCleanUp(self) -> None:

        try:
            if self.o['dry_run']:
                logger.info("deleting queue (dry run) %s" % self.o['queueName'] )
            else:
                logger.info("deleting queue %s" % self.o['queueName'] )
                if hasattr(self,'channel'):
                    self.channel.queue_delete(self.o['queueName'])
        except Exception as err:
            logger.error("failed to {} with {}".format(
                self.o['broker'].url.hostname, err))
            logger.debug('Exception details: ', exc_info=True)

    def newMessages(self) -> list:

        if not self.is_subscriber:  #build_consumer
            logger.error("getting from a publisher")
            return []

        ml = []
        m = self.getNewMessage()
        if m is not None:
            fetched = 1
            ml.append(m)
            while fetched < self.o['batch']:
                m = self.getNewMessage()
                if m is None:
                    break
                ml.append(m)
                fetched += 1

        return ml

    def getNewMessage(self) -> sarracenia.Message:

        if not self.is_subscriber:  #build_consumer
            logger.error("getting from a publisher")
            return None

        try:
            if not self.connection:
                self.getSetup()

            raw_msg = self.channel.basic_get(self.o['queueName'])
            if (raw_msg is None) and (self.connection.connected):
                return None
            else:
                self.metrics['rxByteCount'] += len(raw_msg.body)
                try: 
                    msg = self._msgRawToDict(raw_msg)
                except Exception as err:
                    logger.error("message decode failed. raw message: %s" % raw_msg.body )
                    logger.debug('Exception details: ', exc_info=True)
                    msg = None
                if msg is None:
                    self.metrics['rxBadCount'] += 1
                    return None
                else:
                    self.metrics['rxGoodCount'] += 1
                if hasattr(self.o, 'fixed_headers'):
                    for k in self.o.fixed_headers:
                        msg[k] = self.o.fixed_headers[k]

                logger.debug("new msg: %s" % msg)
                return msg
        except Exception as err:
            logger.warning("failed %s: %s" %
                           (self.o['queueName'], err))
            logger.debug('Exception details: ', exc_info=True)

        if not self.o['message_strategy']['stubborn']:
            return None

        logger.warning('lost connection to broker')
        self.close()
        time.sleep(1)
        return None

    def ack(self, m: sarracenia.Message) -> None:
        """
           do what you need to acknowledge that processing of a message is done.
        """
        if not self.is_subscriber:  #build_consumer
            logger.error("getting from a publisher")
            return

       
        # silent success. retry messages will not have an ack_id, and so will not require acknowledgement.
        if not 'ack_id' in m:
            #logger.warning( f"no ackid present" )
            return

        #logger.info( f"acknowledging {m['ack_id']}" )
        ebo = 1
        while True:
            try:
                if hasattr(self, 'channel'): 
                    self.channel.basic_ack(m['ack_id'])
                del m['ack_id']
                m['_deleteOnPost'].remove('ack_id')
                # Break loop if no exceptions encountered
                return

            except Exception as err:
                logger.warning("failed for tag: %s: %s" % (m['ack_id'], err))
                logger.debug('Exception details: ', exc_info=True)

                # Cleanly close partially broken connection and restablish
                self.close()
                self.getSetup()

            if ebo < 60: ebo *= 2

            logger.info(
                "Sleeping {} seconds before re-trying ack...".format(ebo))
            time.sleep(ebo)

    def putNewMessage(self,
                      message: sarracenia.Message,
                      content_type: str = 'application/json',
                      exchange: str = None ) -> bool:
        """
        put a new message out, to the configured exchange by default.
        """

        if self.is_subscriber:  #build_consumer
            logger.error("publishing from a consumer")
            return False

        # Check connection and channel status, try to reconnect if not connected
        if (not self.connection) or (not self.connection.connected) or (not self.channel.is_open):
            try:
                self.close()
                self.putSetup()
            except Exception as err:
                logger.warning(f"failed, connection was closed/broken and could not be re-opened {exchange}: {err}")
                logger.debug('Exception details: ', exc_info=True)
                return False

        # The caller probably doesn't expect the message to get modified by this method, so use a copy of the message
        body = copy.deepcopy(message)

        version = body['_format']


        if '_deleteOnPost' in body:
            # FIXME: need to delete because building entire JSON object at once.
            # makes this routine alter the message. Ideally, would use incremental
            # method to build json and _deleteOnPost would be a guide of what to skip.
            # library for that is jsonfile, but not present in repos so far.
            for k in body['_deleteOnPost']:
                if k in body:
                    del body[k]
            del body['_deleteOnPost']

        if not exchange:
            if (type(self.o['exchange']) is list):
                if (len(self.o['exchange']) > 1):
                    if ( 'exchangeSplit' in self.o) and self.o['exchangeSplit'] > 1:
                        # FIXME: assert ( len(self.o['exchange']) == self.o['post_exchangeSplit'] )
                        #        if that isn't true... then there is something wrong... should we check ?
                        idx = sum(
                            bytearray(body['identity']['value'],
                                      'ascii')) % len(self.o['exchange'])
                        exchange = self.o['exchange'][idx]
                    else:
                        logger.error(
                            'do not know which exchange to publish to: %s' %
                            self.o['exchange'])
                        return False
                else:
                    exchange = self.o['exchange'][0]
            else:
                exchange = self.o['exchange']

        if self.o['message_ttl']:
            ttl = "%d" * int(
                sarracenia.durationToSeconds(self.o['message_ttl']) * 1000)
        else:
            ttl = "0"

        raw_body, headers, content_type = PostFormat.exportAny( body, version, self.o['topicPrefix'], self.o )

        topic = '.'.join(headers['topic'])
        topic = topic.replace('#', '%23')
        topic = topic.replace('*', '%22')

        if len(topic) >= 255:  # ensure topic is <= 255 characters
            logger.error("message topic too long, truncating")
            mxlen = amqp_ss_maxlen
            while (topic.encode("utf8")[mxlen - 1] & 0xc0 == 0xc0):
                mxlen -= 1
            topic = topic.encode("utf8")[0:mxlen].decode("utf8")

        if self.o['messageDebugDump']:
            logger.info('raw message body: version: %s type: %s %s' %
                             (version, type(raw_body),  raw_body))
            logger.info('raw message headers: type: %s value: %s' % (type(headers),  headers))

        message['post_topic'] = topic
        message['post_exchange'] = exchange
        message['_deleteOnPost'] |= set( ['post_exchange', 'post_topic'] )
        del headers['topic']

        if headers :  
            for k in headers:
                if (type(headers[k]) is str) and (len(headers[k]) >=
                                                      amqp_ss_maxlen):
                    logger.error("message header %s too long, dropping" % k)
                    return False

        AMQP_Message = amqp.Message(raw_body,
                                        content_type=content_type,
                                        application_headers=headers,
                                        expire=ttl,
                                        delivery_mode=2)
        self.metrics['txByteCount'] += len(raw_body) 
        if headers:
            self.metrics['txByteCount'] += len(''.join(str(headers)))

        # timeout option is a float and default is 0.0. basic_publish wants int or None for no timeout
        try:
            if self.o['timeout']:
                pub_timeout = int(self.o['timeout'])
            else:
                pub_timeout = None
        except Exception as err:
            logger.debug('Set pub_timeout to None. Exception details: ', exc_info=True)
            pub_timeout = None

        body=raw_body
        ebo = 1
        try:
            logger.debug( f"trying to publish body: {body} headers: {headers} to {exchange} under: {topic} " )
            self.channel.basic_publish(AMQP_Message, exchange, topic, timeout=pub_timeout)
            # Issue #732: tx_commit can get stuck forever
            self.channel.tx_commit()
            logger.debug("published body: {} headers: {} to {} under: {} ".format(
                          body, headers, exchange, topic))
            self.metrics['txGoodCount'] += 1
            return True  # no failure == success :-)

        except Exception as err:
            logger.warning("failed %s: %s" % (exchange, err))
            logger.debug('Exception details: ', exc_info=True)

            self.metrics['txBadCount'] += 1
            # Issue #466: commenting this out until message_strategy stubborn is working correctly (Issue #537)
            # Always return False when an error occurs and use the DiskQueues to retry, instead of looping. This should
            # eventually be configurable with message_strategy stubborn
            # if True or not self.o['message_strategy']['stubborn']:
            #     return False

            self.close()
            return False # instead of looping

    def close(self) -> None:
        try:
            if self.connection:
                self.connection.collect()
                self.connection.close()

        except Exception as err:
            logger.error("sr_amqp/close 2: {}".format(err))
            logger.debug("sr_amqp/close 2 Exception details:", exc_info=True)
        # FIXME toclose not useful as we don't close channels anymore
        self.metricsDisconnect()
        self.connection = None



"""
Plugin hourtree.py:
    When receiving a file, insert an hourly directory into the local delivery path hierarchy.

Example:
    input A/B/c.gif  --> output A/B/<hour>/c.gif

Usage:
    flowcb sarracenia.flowcb.accept.hourtree.HourTree

"""

import logging
import sys, os, os.path, time, stat
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)

class HourTree(FlowCB):
    def __init__(self, options):
        super().__init__(options,logger)

    def after_accept(self, worklist):
        for message in worklist.incoming:
            datestr = time.strftime('%H', time.localtime())  # pick the hour.
            message['new_dir'] += '/' + datestr  # append the hourly element to the directory tree.

            # insert the additional hourly directory into the path of the file to be written.
            new_fname = message['new_file'].split('/')
            message['new_file'] = '/'.join(new_fname[0:-1]) + '/' + datestr + '/' + new_fname[-1]



"""
Default on_housekeeping handler that:
- Logs Memory and CPU usage.
- Restarts components to deal with memory leaks.

If `MemoryMax` is not in the config, it is automatically calculated with the following procedure:

1. The plugin processes the first `MemoryBaseLineFile` items to reach a steady state.
   - Subscribers process messages-in
   - Posting programs process messages-posted
2. Set `MemoryMax` threshold to `MemoryMultiplier` * (memory usage at time steady state)

If memory use ever exceeds the `MemoryMax` threshold, then the plugin triggers a restart, reducing memory consumption.

Parameters:

MemoryMax : size (default: none)
    Hard coded maximum for tolerable memory consumption.
    Must be suffixed with k/m/g for Kilo/Mega/Giga byte values.
    If not set then the following options will have an effect:

MemoryBaseLineFile : int, optional (default: 100)
    How many files to process before measuring to establish the baseline memory usage.
    (how many files are expected to process before a steady state is reached)

MemoryMultiplier : int, optional (default: 3)
    How many times past the steady state memory footprint you want to allow the component to grow before restarting.
    It could be normal for memory usage to grow, especially if plugins store data in memory.


Returns:
    Nothing, restarts components if memory usage is outside of configured thresholds.
"""

import logging

import os
import psutil
from sarracenia.flowcb import FlowCB
from sarracenia import naturalSize, naturalTime
from sarracenia.featuredetection import features

if features['process']['present']:
    import psutil
else:
    logger.error("missing psutil class cannot monitor process memory usage")


import sys

logger = logging.getLogger(__name__)


class Resources(FlowCB):
    def __init__(self, options):
        super().__init__(options,logger)
        # Set option to neg value to determine if user set in config
        self.o.add_option('MemoryMax', 'size', '0')
        self.o.add_option('MemoryBaseLineFile', 'count', 100)
        self.o.add_option('MemoryMultiplier', 'float', 3)

        self.threshold = None
        ''' Per-process maximum memory footprint that is considered too large, forcing a process restart.'''
        self.transferCount = 0
        self.msgCount = 0

    def on_housekeeping(self):
        if features['process']['present']:
            mem = psutil.Process().memory_info().vms
        else:
            mem = 0

        ost = os.times()
        logger.info(
            f"Current Memory cpu_times: user={ost.user} system={ost.system}")

        # We must set a threshold **after** the config file has been parsed.
        if self.threshold is None:
            # If the config set something, use it.
            if self.o.MemoryMax != 0:
                self.threshold = self.o.MemoryMax

            if self.threshold is None:
                # No user input set, now to figure out what our baseline memory usage is at a steady state
                #   Process MemoryBaseLineFile(s)+ then get a memory reading before setting memory restart threshold.
                if (self.transferCount < self.o.MemoryBaseLineFile) and (
                        self.msgCount < self.o.MemoryBaseLineFile):
                    # Not enough files processed for steady state, continue to wait..
                    logger.info(
                        f"Current mem usage: {naturalSize(mem)}, accumulating count "
                        f"({self.transferCount} or {self.msgCount}/{self.o.MemoryBaseLineFile} so far) "
                        f"before self-setting threshold")
                    return True

                self.threshold = int(self.o.MemoryMultiplier * mem)

            logger.info(
                f"Memory threshold set to: {naturalSize(self.threshold)}"
            )

        logger.info(
            f"Current Memory usage: {naturalSize(mem)} / "
            f"{naturalSize(self.threshold)} = {(mem/self.threshold):.2%}"
        )

        if mem > self.threshold:
            self.restart()
        # self.restart()

        return True

    def restart(self):
        """
        Do an in-place restart of the current process (keeps pid).
        Gets a new memory stack/heap, keeps all file descriptors but replaces the buffers.
        """
        logger.info(
            f"Memory threshold surpassed! Triggering a restart for '{sys.argv}' via '{sys.executable}'"
        )
        # First arg must be the program to be run (absolute path to program)
        # Second arg has to be python for windows, see how this affects the linux side of things..
        # Third arg is the name of the program you wish to run (should be full path to script) plus all the args.
        #   The star unpacks the sys.argv list into the remaining function args

        if sys.platform.startswith(('linux', 'cygwin', 'darwin', 'aix')):
            # Unix* (Linux / Windows/Cygwin / MacOS / AIX) Specific restart
            os.execl(sys.executable, sys.executable, *sys.argv)
        elif sys.platform.startswith('win32'):
            # Windows Specific restart
            os.execl(sys.executable, 'python', *sys.argv)
        else:
            logger.error(
                f'Unknown platform type: "{sys.platform}", attempting default unix process restart..'
            )
            os.execl(sys.executable, sys.executable, *sys.argv)

        # Scream out in agony and die
        logger.critical(
            f'Plugin resources.py:restart() "execl" failed, this should never be logged.'
        )
        exit(1)

    def after_work(self, worklist):
        self.transferCount += len(worklist.ok)
        # if self.threshold is not None:
        #    TODO: Remove this callback when issue #444 is implemented

    def after_accept(self, worklist):
        self.msgCount += len(worklist.incoming)
        # if self.threshold is not None:
        #    TODO: Remove this callback when issue #444 is implemented




import logging
import sarracenia

logger = logging.getLogger(__name__)


class PostFormat:
    """
       A class for controlling the content of notification messages when received or sent,
       as opposed to internal represenation in the Sarracenia.

       Internally All messages are represented as python dictionaries with
       fields identical to v03 messages. 

       PostFormats convert between message payload protocols and in-memory Sarracenia.Message
       formats.

       To add new post formats just add sub-classes. they will be discovered at run-time.

       Every message read in will have include the *_format* field, identifying the format
       of the incoming message.

       each subclass implements:

       def mine(payload, headers, content_type) -> bool:

       which returns True of the message is in the given post format.

       subclasses have encode and decode routines with different signatures.

       def import(payload, headers, topic, topicPrefix) -> sarracenia.Message:
            given this wire payload, return a native data structure (v03 in a dictionary)

       def export(msg) -> ( body (str), headers (dict), contenty_typ (str) ) :
            opposite of decode, take a v03 message and encode it for sending on the wire.
            returns a tuple to be passed to mqp apis.

       These are used by the Any routines in this class.
    """

    def content_type(post_format):
        for sc in PostFormat.__subclasses__():
            if post_format == sc.__name__.lower():
                return sc.content_type() 
        return None

        return self.mimetype

    @staticmethod
    def importAny(payload, headers, content_type, options ) -> sarracenia.Message:
        """
          given a message in a wire format, with the given properties (or headers) in a dictionary,
          return the message as a normalized v03 message.

          msg['_format'] will be set to the name of the postformat detected and decoded.
       """
        for sc in PostFormat.__subclasses__():
            #logger.info( f" sc={sc}, scct={sc.content_type()}, content_type={content_type} " )
            if sc.mine(payload, headers, content_type, options ):
                return sc.importMine(payload, headers, options )
        return None

        pass

    @staticmethod
    def exportAny(msg, post_format='v03', topicPrefix=[ 'v03' ], options={ 'post_format': 'v03', 'topicPrefix':'v03' } ) -> (str, dict, str):
        """
          return a tuple of the encoded message body, a headers dict, and content_type
          and a completed topic as a list as one header.
       """
        for sc in PostFormat.__subclasses__():
            if post_format == sc.__name__.lower():
                return sc.exportMine( msg, options ) 

        return None, None, None

    def topicDerive(msg, options ) -> list:
        """
           Sarracenia standard topic derivation.

           https://metpx.github.io/sarracenia/Explanation/Concepts.html#amqp-v09-rabbitmq-settings

        """

        if options['broker'].url.scheme.startswith('mqtt'):
            if ( 'exchange' in options ) and ( 'topicPrefix' in options ):
                if 'exchangeSplit' in options and options['exchangeSplit'] > 1:
                    idx = sum( bytearray(msg['identity']['value'], 'ascii')) % len(options['exchange'])
                    exchange = options['exchange'][idx]
                else:
                    exchange = options['exchange'][0]
            topic_prefix = [exchange] + options['topicPrefix']
            topic_separator='/'
        else:
            topic_prefix = options['topicPrefix']
            topic_separator='.'

        if 'topic' in msg:
            if type(msg['topic']) is list:
                topic = msg['topic']
            else:
                topic = msg['topic'].split(topic_separator)
        elif 'topic' in options and options['topic'] and (type(options['topic']) is not list):
            topic = options['topic'].split(topic_separator)
        else:
            if 'relPath' in msg: 
                topic = topic_prefix + msg['relPath'].split('/')[0:-1]
            elif 'subtopic' in msg:
                topic = topic_prefix + msg['subtopic']  
            else:
                topic = topic_prefix
        return topic

   

# test for v04 first, because v03 may claim all other JSON.
import sarracenia.postformat.wis
import sarracenia.postformat.v03
import sarracenia.postformat.v02




"""
Plugin renamedmf.py:
    This is an  example of the usage of after_accept script
    to rename the product when it is downloaded
    It adds a ':datetime' to the name of the product
    unlikely to be useful to others, except as example.

Example:
    takes px name     : CACN00_CWAO_081900__WZS_14623:pxatx:CWAO:CA:5:Direct:20081008190602
    add datetimestamp : CACN00_CWAO_081900__WZS_14623:pxatx:CWAO:CA:5:Direct:20081008190602:20081008190612

Usage:
    flowcb sarracenia.flowcb.accept.renamedmf.RenameDMF

"""
import logging
import time
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)

class RenameDMF(FlowCB):
    def __init__(self, options):
        super().__init__(options,logger)

    def after_accept(self, worklist):
        for message in worklist.incoming:
            datestr = time.strftime(':%Y%m%d%H%M%S', time.localtime())
            message['new_file'] += datestr
            message['rename'] += datestr



"""
Plugin tolocal.py:
    This is a helper script to work with filters.
    What a data pump advertises, it will usually use Web URL, but if one is
    on a server where the files are available, it is more efficient to access 
    them as local files, so filters operate on file urls.  
    
Example:
    baseDir /var/www/html
    url is http://localhost/<date>/<src>/input/file.txt

    flowcb sarracenia.flowcb.accept.tolocal.ToLocal   # converts web URL to file URL

            http://localhost/ --> file://var/www/html/
            url is now file://var/www/html/<date>/<src>/input/file.txt
            m.savedurl = http://localhost/

    flowcb sarracenia.flowcb.accept.<some converter that works on local files.>

            A new file is created in another directory.
            url is now file://var/www/<date>/<src>/output/file.txt

    flowcb sarracenia.flowcb.accept.tohttp.ToHttp     # turns the file URL back into a web one.

            file://var/www/html/ --> http:///localhost/
            url is now:   http://localhost/<date>/<src>/output/file.txt


    The regular expression used to find the web url matches either http or https
    and just captures upto the first '/'.

    if you need to capture a different kind of url, such as ...

    https://hostname/~user/ ....

    The easiest way is to set toLocalUrl as follows:

    baseDir /home/user/www
    toLocalUrl (https://hostname/~user/)

    the parentheses around the URL set the value of to be put in m.savedurl that
    will be restored when the companion plugin msg_2http is called.

Usage: 
    flowcb sarracenia.flowcb.accept.tolocal.ToLocal
"""
import logging
import re
from sarracenia.flowcb import FlowCB
logger = logging.getLogger(__name__)


class ToLocal(FlowCB):
    def __init__(self, options):
        super().__init__(options,logger)

        self._ldocroot = None

        if self.o.baseDir:
            self._ldocroot = self.o.baseDir

        self.o.add_option('toLocalRoot', 'str')
        if self.o.toLocalRoot:
            self._ldocroot = self.o.toLocalRoot

        self._lurlre = re.compile("(http[s]{0,1}://[^/]+/)")

        self.o.add_option('toLocalUrl', 'str')
        if self.o.toLocalUrl:
            self._lurlre = re.compile(self.o.toLocalUrl)

    def after_accept(self, worklist):
        for message in worklist.incoming:
            # TODO should all these be logger.error? should we append
            #  to worklist.rejected or worklist.failed at some point?
            logger.debug("input: urlstr: %s" % message['urlstr'])

            message['savedurl'] = self._lurlre.match(message['urlstr']).group(1)
            message['urlstr'] = 'file:/%s' % self._lurlre.sub(self._ldocroot + '/', message['urlstr'])

            logger.debug("doc_root=%s " % (self.o.baseDir))
            logger.debug("output: savedurl: %s" % (message['savedurl']))



"""
  This plugin delays processing of messages by *message_delay* seconds

  sarracenia.flowcb.msg.fdelay 30
  import sarracenia.flowcb.filter.fdelay.Fdelay

  or more simply:

  fdelay 30
  callback filter.fdelay

  every message will be at least 30 seconds old before it is forwarded by this plugin.
  in the meantime, the message is placed on the retry queue by marking it as failed.

"""
import logging
import os
import os.path
import stat

from sarracenia.flowcb import FlowCB
from sarracenia import timestr2flt, nowflt

logger = logging.getLogger(__name__)


class Fdelay(FlowCB):
    def __init__(self, options):

        super().__init__(options,logger)

        logging.basicConfig(format=self.o.logFormat,
                            level=getattr(logging, self.o.logLevel.upper()))

        self.o.add_option('msg_fdelay', 'duration', 60)
        self.o.add_option('fdelay', 'duration', 60)

        #parent.declare_option('fdelay')
        if hasattr(self.o, 'msg_fdelay'):
            self.o.fdelay = self.o.msg_fdelay


    def after_accept(self, worklist):
        # Prepare msg delay test
        outgoing = []
        for m in worklist.incoming:
            # Test msg delay
            elapsedtime = nowflt() - timestr2flt(m['pubTime'])

            if 'fileOp' in m and 'remove' in m['fileOp'] :
                # 'remove' msg will be removed by itself
                worklist.rejected.append(m)
                logger.debug('marked rejected 0 (file removal)')
                continue

            # Test msg delay
            elapsedtime = nowflt() - timestr2flt(m['pubTime'])
            if elapsedtime < self.o.fdelay:
                dbg_msg = "message not old enough, sleeping for {:.3f} seconds"
                logger.debug(
                    dbg_msg.format(elapsedtime, self.o.fdelay - elapsedtime))
                worklist.failed.append(m)
                logger.debug('marked failed 1 (message not old enough)')
                continue

            # Prepare file delay test
            if '/cfr/' in m['new_dir']:
                f = os.path.join(m['new_dir'], m['new_file'])
            else:
                f = '/' + m['relPath']
            if not os.path.exists(f):
                logger.debug("did not find file {}".format(f))
                worklist.failed.append(m)
                logger.debug('marked failed 2 (file not found)')
                continue

            # Test file delay
            filetime = os.path.getmtime(f)
            elapsedtime = nowflt() - filetime
            if elapsedtime < self.o.fdelay:
                dbg_msg = "file not old enough, sleeping for {:.3f} seconds"
                logger.debug(
                    dbg_msg.format(elapsedtime, self.o.fdelay - elapsedtime))
                worklist.failed.append(m)
                logger.debug('marked failed 3 file not old enough')
                continue

            logger.debug('appending to outgoing')
            outgoing.append(m)

        worklist.incoming = outgoing



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) His Majesty The King in Right of Canada, Environment Canada, 2008-2020
#

"""
Description:
    AM Server         

    This is a sr3 plugin built to migrate the reception of ECCC's
    proprietary Alpha Manager(AM) socket protocol.

    For more information on the origins of AM and Sundew visit https://github.com/MetPX/Sundew/blob/main/doc/historical/Origins.rst

    Code overview:
        To start, when on_start is called, the socket binds, listens and afterwards accepts a connection when one is received from a remote host.
        The receiver acts like an amtcp server. Instances are split into child/parent once an outbound connection is accepted.
        The child proceeds to the receiver service while the parent stays and keeps accepting connections.   
        In the service, bulletins are received and are stored locally inside of a given filepath.

    Options:
        AllowIPs (list): Filters outbound connections with provided IPs. All other IPs won't be accepted. 
        If unselected, accept all IPs. 

        directory (string): Specifies the directory where the bulletin files are to be stored. 

    NOTE: AM protocol cannot correct data corruption and cannot be stopped without data loss.
    Precautions should be taken during maintenance interventions.

Platform:
    Linux/Mac: because of process forking, this will not work on Windows. 

Usage:
    flowcb amserver
    See sarracenia/examples/flow/amserver.conf for an example config file.

Author:
    André LeBlanc, ANL, Autumn 2022
"""

import logging, socket, struct, time, sys, os, signal, ipaddress
import urllib.parse
import sarracenia
import sarracenia.config
from sarracenia.flowcb import FlowCB
from random import randint

logger = logging.getLogger(__name__)

class Amserver(FlowCB):

    def __init__(self, options):
        
        super().__init__(options,logger)

        self.url = urllib.parse.urlparse(self.o.sendTo)

        self.inBuffer = bytes()
        self.limit = 32678
        self.patternAM = '80sII4siIII20s'
        self.sizeAM = struct.calcsize(self.patternAM)
        self.o.add_option('AllowIPs', 'list', [])
        self.host = self.url.netloc.split(':')[0]
        self.port = int(self.url.netloc.split(':')[1])
        self.minnum = 00000
        self.maxnum = 99999 
        self.remoteHost = None

        # Initialise socket
        ## Create a TCP socket
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        # Add signal handler
        ## Override outer signal handler with a default one to exit correctly.
        signal.signal(signal.SIGTERM, signal.SIG_DFL)

 
    def __WaitForRemoteConnection__(self):

        if self.host == 'None':
            raise Exception("No host was specified. Exiting.")

        try:
            # Bind socket to specified host and listen
            self.s.bind((self.host, self.port)) 
            self.s.listen(1)
            logger.info("Socket listening on host %s and port %d.", self.host, self.port)
        except socket.error as e:
                logger.error(f"Bind failed. Retrying. Error message: {e.args}")
                time.sleep(5)
        
        child_inst = 2

        while True:
            
                try:
                    # Accept the connection from socket
                    logger.info("Trying to accept connection.")

                    try:
                        conn, self.remoteHost = self.s.accept()
                        time.sleep(1)

                    except Exception as e:
                        logger.error(f"Stopping accept. Exiting. Error message: {e}")
                        sys.exit(0)   

                    if self.o.AllowIPs:
                        if self.remoteHost[0] not in self.o.AllowIPs:
                            logger.debug(f"Connection to IP {self.remoteHost[0]} rejected. Not a part of the Accepted IPs list.")
                            conn.close()
                            continue
                    
                    # Instance forks
                    ## Instance 1 (Parent, pid=child_pid): Stays in the loop trying to accept other connections. 
                    ## Instance 2 (Child, pid=0): Exits loop. Proceeds to initialise the service with the remote host.
                    pid = os.fork()

                    if pid == 0:
                        ## Close the unconnected socket instance as it is unused in the service.
                        self.s.close()

                        ## Set the logfiles properly
                        sarracenia.config.cfglogs(self.o, self.o.component, self.o.config, self.o.logLevel, child_inst)

                        self.o.no = child_inst
                        logger.info(f"Starting up service with host {self.remoteHost[0]}")
                        break                    

                    elif pid == -1:
                        raise logger.exception("Connection could not fork. Exiting.")
        
                    else:
                        pidfilename = sarracenia.config.get_pid_filename(
                        None, self.o.component, self.o.config, child_inst)

                        with open(pidfilename, 'w') as pfn:
                            pfn.write('%d' % pid)

                        ## Close the connected socket instance as it is unused in the parent
                        conn.close()
                        logger.info(f"Forked child from host {self.remoteHost[0]} with instance number {child_inst} and pid {pid}")

                        child_inst += 1
                        pass
                   
                except Exception:
                    logger.error(f"Couldn't accept connection. Parent or child failed. Retrying to accept.")
                    # self.s.close()
                    # conn.close()
                    time.sleep(1)
                
        logger.info("Connection accepted with IP %s on port %d. Starting service.", self.remoteHost[0], self.port)     

        return conn                       


    def on_start(self):
        # Set ipadresses in proper format
        for IP in self.o.AllowIPs:
            IP = ipaddress.ip_address(IP)

        # If there are remaining instances, delete their filepaths and exit.
        if self.o.no != 1:
            pidfilename = sarracenia.config.get_pid_filename(None, self.o.component, self.o.config, self.o.no)
            if os.path.exists(pidfilename):
                os.unlink(pidfilename)
            sys.exit(0)
        
        self.conn = self.__WaitForRemoteConnection__()
    

    def on_stop(self):
        logger.info("On stop called. Exiting.")

        # Close all socket connections (from parent and children) and exit.
        self.s.close()
        if self.o.no == 1:
            pass
        else:
            self.conn.close()
        sys.exit(0)
        

    def AddBuffer(self):
        # try:
        try:
            tmp = self.conn.recv(self.limit)

        except Exception as e:
            tmp = ''
            logger.error(f"Reception has been interrupted. Closing connection and exiting. Error message: {e}")

        if tmp == '':
            self.conn.close()
            raise Exception()
        
        self.inBuffer = self.inBuffer + tmp

        # except socket.error:
            # (type, value, tb) = sys.exc_info()
            # logger.warning("Type: %s, Value: %s, [socket.recv(%d)]" % (type, value, self.limit))
            
            
    def CheckNextMsgStatus(self):

        # Only unpack data if a bulletin is received
        ## When unpacking, the length of the header is vital since it allows the receiver to extract the bulletin contents from the buffer.
        ## AND it determines if all the bulletin contents are available inside the buffer.
        if len(self.inBuffer) >= self.sizeAM:
            (header, src_inet, dst_inet, threads, start, length, firsttime, timestamp, future) = \
                    struct.unpack(self.patternAM,self.inBuffer[0:self.sizeAM])
        else:
            return 'INCOMPLETE'

        length = socket.ntohl(length)

        # logger.debug(f"Buffer contents: {self.inBuffer}")
        
        if len(self.inBuffer) >= self.sizeAM + length:
            return 'OK'
        else:
            return 'INCOMPLETE'


    def unwrapmsg(self):

        status = self.CheckNextMsgStatus()

        if status == 'OK':
            (self.header,src_inet,dst_inet,threads,start,length,firsttime,timestamp,future) = \
                    struct.unpack(self.patternAM,self.inBuffer[0:self.sizeAM])

            length = socket.ntohl(length)

            bulletin = self.inBuffer[self.sizeAM:self.sizeAM + length]
            longlen = self.sizeAM + length

            logger.debug("Gather successful.")

            return (bulletin, longlen)
        
        else:
            return '', 0


    def gather(self):

        self.AddBuffer()

        newmsg = []

        while True:
            status = self.CheckNextMsgStatus()
            
            if status == 'INCOMPLETE':
                break

            if status == 'OK':
                (bulletin, longlen) = self.unwrapmsg()

                # Set buffer for next bulletin ingestion
                self.inBuffer = self.inBuffer[longlen:]

                logger.debug(f"Bulletin contents: {bulletin}")

                parse = self.header.split(b'\0',1)
                bulletinHeader = parse[0].decode('iso-8859-1').replace(' ', '_')

                # Create a file for new messages and let sarracenia format the data
                try:
                    ## Filenames have the following naming scheme:
                    ##   1. Bulletin header
                    ##   2. Counter (makes filename unique for each bulletin)
                    ##   IF A* or R* present in header, include in filename
                    ##
                    ##   Example: SXVX65_KWNB_181800_RRB__99705
                    filepath = self.o.directory + os.sep + bulletinHeader + '__' +  f"{randint(self.minnum, self.maxnum)}".zfill(len(str(self.maxnum)))

                    file = open(filepath, 'wb')
                    file.write(bulletin)
                    file.close()
                    st = os.stat(filepath)

                    sarramsg = sarracenia.Message.fromFileData(filepath, self.o, lstat=st)
                    newmsg.append(sarramsg)

                except:
                    logger.error("Unable to generate bulletin file.")

        return newmsg 



"""
    Example of consuming data from a Sarracenia Data Pump.

"""
import sarracenia.moth
import sarracenia.moth.amqp
import sarracenia.credentials

import time
import socket
import urllib.request
import xml.etree.ElementTree as ET


options = sarracenia.moth.default_options
options.update(sarracenia.moth.amqp.default_options)
options['broker'] = sarracenia.credentials.Credential(
    'amqps://anonymous:anonymous@hpfx.collab.science.gc.ca')
options['topicPrefix'] = ['v02', 'post']
options['bindings'] = [('xpublic', ['v02', 'post'],
                        ['*', 'WXO-DD', 'observations', 'swob-ml', '#'])]
options['queueName'] = 'q_anonymous_' + socket.getfqdn(
) + '_SomethingHelpfulToYou'

#print('options: %s' % options)

h = sarracenia.moth.Moth.subFactory(options)

count = 0

while count < 10:
    messages = h.newMessages()
    for m in messages:
        dataUrl = m['baseUrl']
        if 'retreivePath' in m:
            dataUrl += m['retreivePath']
        else:
            dataUrl += m['relPath']

        print("url %d: %s" % (count, dataUrl))
        with urllib.request.urlopen(dataUrl) as f:
            vxml = f.read().decode('utf-8')
            xmlData = ET.fromstring(vxml)

            stn_name = ''
            tc_id = ''
            lat = ''
            lon = ''
            air_temp = ''

            for i in xmlData.iter():
                name = i.get('name')
                if name == 'stn_nam':
                    stn_name = i.get('value')
                elif name == 'tc_id':
                    tc_id = i.get('value')
                elif name == 'lat':
                    lat = i.get('value')
                elif name == 'long':
                    lon = i.get('value')
                elif name == 'air_temp':
                    air_temp = i.get('value')

            print('station: %s, tc_id: %s, lat: %s, long: %s, air_temp: %s' %
                  (stn_name, tc_id, lat, lon, air_temp))
        h.ack(m)
        count += 1
        if count > 10:
            break
    time.sleep(1)

h.cleanup()  # remove server-side queue defined by Factory.
h.close()
print("obtained 10 product announcements")
